@@ -1,4 +1,3 @@
-import select
 import socket
 import threading
 import time
@@ -7,7 +6,6 @@ from types import TracebackType
 from typing import Optional, Type
 
 from .._exceptions import (
-    CloseError,
     ConnectError,
     ConnectTimeout,
     ReadError,
@@ -17,6 +15,7 @@ from .._exceptions import (
     map_exceptions,
 )
 from .._types import TimeoutDict
+from .._utils import is_socket_readable
 
 
 class SyncSocketStream:
@@ -74,12 +73,13 @@ class SyncSocketStream:
 
     def close(self) -> None:
         with self.write_lock:
-            with map_exceptions({socket.error: CloseError}):
+            try:
                 self.sock.close()
+            except socket.error:
+                pass
 
-    def is_connection_dropped(self) -> bool:
-        rready, _wready, _xready = select.select([self.sock], [], [], 0)
-        return bool(rready)
+    def is_readable(self) -> bool:
+        return is_socket_readable(self.sock)
 
 
 class SyncLock:
@@ -173,3 +173,6 @@ class SyncBackend:
 
     def time(self) -> float:
         return time.monotonic()
+
+    def sleep(self, seconds: float) -> None:
+        time.sleep(seconds)
