@@ -5,12 +5,7 @@ from .._backends.sync import SyncBackend, SyncLock, SyncSocketStream, SyncBacken
 from .._exceptions import ConnectError, ConnectTimeout
 from .._types import URL, Headers, Origin, TimeoutDict
 from .._utils import exponential_backoff, get_logger, url_to_origin
-from .base import (
-    SyncByteStream,
-    SyncHTTPTransport,
-    ConnectionState,
-    NewConnectionRequired,
-)
+from .base import SyncByteStream, SyncHTTPTransport, NewConnectionRequired
 from .http import SyncBaseHTTPConnection
 from .http11 import SyncHTTP11Connection
 
@@ -25,6 +20,7 @@ class SyncHTTPConnection(SyncHTTPTransport):
         origin: Origin,
         http1: bool = True,
         http2: bool = False,
+        keepalive_expiry: float = None,
         uds: str = None,
         ssl_context: SSLContext = None,
         socket: SyncSocketStream = None,
@@ -33,13 +29,14 @@ class SyncHTTPConnection(SyncHTTPTransport):
         backend: SyncBackend = None,
     ):
         self.origin = origin
-        self.http1 = http1
-        self.http2 = http2
-        self.uds = uds
-        self.ssl_context = SSLContext() if ssl_context is None else ssl_context
+        self._http1_enabled = http1
+        self._http2_enabled = http2
+        self._keepalive_expiry = keepalive_expiry
+        self._uds = uds
+        self._ssl_context = SSLContext() if ssl_context is None else ssl_context
         self.socket = socket
-        self.local_address = local_address
-        self.retries = retries
+        self._local_address = local_address
+        self._retries = retries
 
         alpn_protocols: List[str] = []
         if http1:
@@ -47,36 +44,74 @@ class SyncHTTPConnection(SyncHTTPTransport):
         if http2:
             alpn_protocols.append("h2")
 
-        self.ssl_context.set_alpn_protocols(alpn_protocols)
+        self._ssl_context.set_alpn_protocols(alpn_protocols)
 
         self.connection: Optional[SyncBaseHTTPConnection] = None
-        self.is_http11 = False
-        self.is_http2 = False
-        self.connect_failed = False
-        self.expires_at: Optional[float] = None
-        self.backend = SyncBackend() if backend is None else backend
+        self._is_http11 = False
+        self._is_http2 = False
+        self._connect_failed = False
+        self._expires_at: Optional[float] = None
+        self._backend = SyncBackend() if backend is None else backend
 
     def __repr__(self) -> str:
-        http_version = "UNKNOWN"
-        if self.is_http11:
-            http_version = "HTTP/1.1"
-        elif self.is_http2:
-            http_version = "HTTP/2"
-        return f"<SyncHTTPConnection http_version={http_version} state={self.state}>"
+        return f"<SyncHTTPConnection [{self.info()}]>"
 
     def info(self) -> str:
         if self.connection is None:
-            return "Not connected"
-        elif self.state == ConnectionState.PENDING:
-            return "Connecting"
+            return "Connection failed" if self._connect_failed else "Connecting"
         return self.connection.info()
 
+    def should_close(self) -> bool:
+        """
+        Return `True` if the connection is in a state where it should be closed.
+        This occurs when any of the following occur:
+
+        * There are no active requests on an HTTP/1.1 connection, and the underlying
+          socket is readable. The only valid state the socket can be readable in
+          if this occurs is when the b"" EOF marker is about to be returned,
+          indicating a server disconnect.
+        * There are no active requests being made and the keepalive timeout has passed.
+        """
+        if self.connection is None:
+            return False
+        return self.connection.should_close()
+
+    def is_idle(self) -> bool:
+        """
+        Return `True` if the connection is currently idle.
+        """
+        if self.connection is None:
+            return False
+        return self.connection.is_idle()
+
+    def is_closed(self) -> bool:
+        if self.connection is None:
+            return self._connect_failed
+        return self.connection.is_closed()
+
+    def is_available(self) -> bool:
+        """
+        Return `True` if the connection is currently able to accept an outgoing request.
+        This occurs when any of the following occur:
+
+        * The connection has not yet been opened, and HTTP/2 support is enabled.
+          We don't *know* at this point if we'll end up on an HTTP/2 connection or
+          not, but we *might* do, so we indicate availability.
+        * The connection has been opened, and is currently idle.
+        * The connection is open, and is an HTTP/2 connection. The connection must
+          also not currently be exceeding the maximum number of allowable concurrent
+          streams and must not have exhausted the maximum total number of stream IDs.
+        """
+        if self.connection is None:
+            return self._http2_enabled and not self.is_closed
+        return self.connection.is_available()
+
     @property
     def request_lock(self) -> SyncLock:
         # We do this lazily, to make sure backend autodetection always
         # runs within an async context.
         if not hasattr(self, "_request_lock"):
-            self._request_lock = self.backend.create_lock()
+            self._request_lock = self._backend.create_lock()
         return self._request_lock
 
     def handle_request(
@@ -91,18 +126,16 @@ class SyncHTTPConnection(SyncHTTPTransport):
         timeout = cast(TimeoutDict, extensions.get("timeout", {}))
 
         with self.request_lock:
-            if self.state == ConnectionState.PENDING:
+            if self.connection is None:
+                if self._connect_failed:
+                    raise NewConnectionRequired()
                 if not self.socket:
                     logger.trace(
                         "open_socket origin=%r timeout=%r", self.origin, timeout
                     )
                     self.socket = self._open_socket(timeout)
                 self._create_connection(self.socket)
-            elif self.state in (ConnectionState.READY, ConnectionState.IDLE):
-                pass
-            elif self.state == ConnectionState.ACTIVE and self.is_http2:
-                pass
-            else:
+            elif not self.connection.is_available():
                 raise NewConnectionRequired()
 
         assert self.connection is not None
@@ -119,34 +152,34 @@ class SyncHTTPConnection(SyncHTTPTransport):
     def _open_socket(self, timeout: TimeoutDict = None) -> SyncSocketStream:
         scheme, hostname, port = self.origin
         timeout = {} if timeout is None else timeout
-        ssl_context = self.ssl_context if scheme == b"https" else None
+        ssl_context = self._ssl_context if scheme == b"https" else None
 
-        retries_left = self.retries
+        retries_left = self._retries
         delays = exponential_backoff(factor=RETRIES_BACKOFF_FACTOR)
 
         while True:
             try:
-                if self.uds is None:
-                    return self.backend.open_tcp_stream(
+                if self._uds is None:
+                    return self._backend.open_tcp_stream(
                         hostname,
                         port,
                         ssl_context,
                         timeout,
-                        local_address=self.local_address,
+                        local_address=self._local_address,
                     )
                 else:
-                    return self.backend.open_uds_stream(
-                        self.uds, hostname, ssl_context, timeout
+                    return self._backend.open_uds_stream(
+                        self._uds, hostname, ssl_context, timeout
                     )
             except (ConnectError, ConnectTimeout):
                 if retries_left <= 0:
-                    self.connect_failed = True
+                    self._connect_failed = True
                     raise
                 retries_left -= 1
                 delay = next(delays)
-                self.backend.sleep(delay)
+                self._backend.sleep(delay)
             except Exception:  # noqa: PIE786
-                self.connect_failed = True
+                self._connect_failed = True
                 raise
 
     def _create_connection(self, socket: SyncSocketStream) -> None:
@@ -154,38 +187,31 @@ class SyncHTTPConnection(SyncHTTPTransport):
         logger.trace(
             "create_connection socket=%r http_version=%r", socket, http_version
         )
-        if http_version == "HTTP/2" or (self.http2 and not self.http1):
+        if http_version == "HTTP/2" or (
+            self._http2_enabled and not self._http1_enabled
+        ):
             from .http2 import SyncHTTP2Connection
 
-            self.is_http2 = True
+            self._is_http2 = True
             self.connection = SyncHTTP2Connection(
-                socket=socket, backend=self.backend, ssl_context=self.ssl_context
+                socket=socket,
+                keepalive_expiry=self._keepalive_expiry,
+                backend=self._backend,
             )
         else:
-            self.is_http11 = True
+            self._is_http11 = True
             self.connection = SyncHTTP11Connection(
-                socket=socket, ssl_context=self.ssl_context
+                socket=socket, keepalive_expiry=self._keepalive_expiry
             )
 
-    @property
-    def state(self) -> ConnectionState:
-        if self.connect_failed:
-            return ConnectionState.CLOSED
-        elif self.connection is None:
-            return ConnectionState.PENDING
-        return self.connection.get_state()
-
-    def is_socket_readable(self) -> bool:
-        return self.connection is not None and self.connection.is_socket_readable()
-
-    def mark_as_ready(self) -> None:
-        if self.connection is not None:
-            self.connection.mark_as_ready()
-
-    def start_tls(self, hostname: bytes, timeout: TimeoutDict = None) -> None:
+    def start_tls(
+        self, hostname: bytes, ssl_context: SSLContext, timeout: TimeoutDict = None
+    ) -> None:
         if self.connection is not None:
             logger.trace("start_tls hostname=%r timeout=%r", hostname, timeout)
-            self.socket = self.connection.start_tls(hostname, timeout)
+            self.socket = self.connection.start_tls(
+                hostname, ssl_context, timeout
+            )
             logger.trace("start_tls complete hostname=%r timeout=%r", hostname, timeout)
 
     def close(self) -> None:
