@@ -1,5 +1,7 @@
+import enum
+import time
 from ssl import SSLContext
-from typing import Iterator, List, Tuple, Union, cast
+from typing import Iterator, List, Optional, Tuple, Union, cast
 
 import h11
 
@@ -8,7 +10,7 @@ from .._bytestreams import IteratorByteStream
 from .._exceptions import LocalProtocolError, RemoteProtocolError, map_exceptions
 from .._types import URL, Headers, TimeoutDict
 from .._utils import get_logger
-from .base import SyncByteStream, ConnectionState
+from .base import SyncByteStream, NewConnectionRequired
 from .http import SyncBaseHTTPConnection
 
 H11Event = Union[
@@ -20,32 +22,79 @@ H11Event = Union[
     h11.ConnectionClosed,
 ]
 
+
+class ConnectionState(enum.IntEnum):
+    NEW = 0
+    ACTIVE = 1
+    IDLE = 2
+    CLOSED = 3
+
+
 logger = get_logger(__name__)
 
 
 class SyncHTTP11Connection(SyncBaseHTTPConnection):
     READ_NUM_BYTES = 64 * 1024
 
-    def __init__(self, socket: SyncSocketStream, ssl_context: SSLContext = None):
+    def __init__(self, socket: SyncSocketStream, keepalive_expiry: float = None):
         self.socket = socket
-        self.ssl_context = SSLContext() if ssl_context is None else ssl_context
 
-        self.h11_state = h11.Connection(our_role=h11.CLIENT)
-
-        self.state = ConnectionState.ACTIVE
+        self._keepalive_expiry: Optional[float] = keepalive_expiry
+        self._should_expire_at: Optional[float] = None
+        self._h11_state = h11.Connection(our_role=h11.CLIENT)
+        self._state = ConnectionState.NEW
 
     def __repr__(self) -> str:
-        return f"<SyncHTTP11Connection state={self.state}>"
+        return f"<SyncHTTP11Connection [{self._state.name}]>"
+
+    def _now(self) -> float:
+        return time.monotonic()
+
+    def _server_disconnected(self) -> bool:
+        """
+        Return True if the connection is idle, and the underlying socket is readable.
+        The only valid state the socket can be readable here is when the b""
+        EOF marker is about to be returned, indicating a server disconnect.
+        """
+        return self._state == ConnectionState.IDLE and self.socket.is_readable()
+
+    def _keepalive_expired(self) -> bool:
+        """
+        Return True if the connection is idle, and has passed it's keepalive
+        expiry time.
+        """
+        return (
+            self._state == ConnectionState.IDLE
+            and self._should_expire_at is not None
+            and self._now() >= self._should_expire_at
+        )
 
     def info(self) -> str:
-        return f"HTTP/1.1, {self.state.name}"
+        return f"HTTP/1.1, {self._state.name}"
 
-    def get_state(self) -> ConnectionState:
-        return self.state
+    def should_close(self) -> bool:
+        """
+        Return `True` if the connection is in a state where it should be closed.
+        """
+        return self._server_disconnected() or self._keepalive_expired()
 
-    def mark_as_ready(self) -> None:
-        if self.state == ConnectionState.IDLE:
-            self.state = ConnectionState.READY
+    def is_idle(self) -> bool:
+        """
+        Return `True` if the connection is currently idle.
+        """
+        return self._state == ConnectionState.IDLE
+
+    def is_closed(self) -> bool:
+        """
+        Return `True` if the connection has been closed.
+        """
+        return self._state == ConnectionState.CLOSED
+
+    def is_available(self) -> bool:
+        """
+        Return `True` if the connection is currently able to accept an outgoing request.
+        """
+        return self._state == ConnectionState.IDLE
 
     def handle_request(
         self,
@@ -55,9 +104,19 @@ class SyncHTTP11Connection(SyncBaseHTTPConnection):
         stream: SyncByteStream,
         extensions: dict,
     ) -> Tuple[int, Headers, SyncByteStream, dict]:
+        """
+        Send a single HTTP/1.1 request.
+
+        Note that there is no kind of task/thread locking at this layer of interface.
+        Dealing with locking for concurrency is handled by the `SyncHTTPConnection`.
+        """
         timeout = cast(TimeoutDict, extensions.get("timeout", {}))
 
-        self.state = ConnectionState.ACTIVE
+        if self._state in (ConnectionState.NEW, ConnectionState.IDLE):
+            self._state = ConnectionState.ACTIVE
+            self._should_expire_at = None
+        else:
+            raise NewConnectionRequired()
 
         self._send_request(method, url, headers, timeout)
         self._send_request_body(stream, timeout)
@@ -78,10 +137,10 @@ class SyncHTTP11Connection(SyncBaseHTTPConnection):
         return (status_code, headers, response_stream, extensions)
 
     def start_tls(
-        self, hostname: bytes, timeout: TimeoutDict = None
+        self, hostname: bytes, ssl_context: SSLContext, timeout: TimeoutDict = None
     ) -> SyncSocketStream:
         timeout = {} if timeout is None else timeout
-        self.socket = self.socket.start_tls(hostname, self.ssl_context, timeout)
+        self.socket = self.socket.start_tls(hostname, ssl_context, timeout)
         return self.socket
 
     def _send_request(
@@ -117,7 +176,7 @@ class SyncHTTP11Connection(SyncBaseHTTPConnection):
         Send a single `h11` event to the network, waiting for the data to
         drain before returning.
         """
-        bytes_to_send = self.h11_state.send(event)
+        bytes_to_send = self._h11_state.send(event)
         self.socket.write(bytes_to_send, timeout)
 
     def _receive_response(
@@ -160,7 +219,7 @@ class SyncHTTP11Connection(SyncBaseHTTPConnection):
         """
         while True:
             with map_exceptions({h11.RemoteProtocolError: RemoteProtocolError}):
-                event = self.h11_state.next_event()
+                event = self._h11_state.next_event()
 
             if event is h11.NEED_DATA:
                 data = self.socket.read(self.READ_NUM_BYTES, timeout)
@@ -171,13 +230,12 @@ class SyncHTTP11Connection(SyncBaseHTTPConnection):
                 #     ConnectionClosed when role=SERVER and state=SEND_RESPONSE
                 #
                 # Which is accurate, but not very informative from an end-user
-                # perspective. Instead we handle this case distinctly and treat
-                # it as a ConnectError.
-                if data == b"" and self.h11_state.their_state == h11.SEND_RESPONSE:
+                # perspective. Instead we handle messaging for this case distinctly.
+                if data == b"" and self._h11_state.their_state == h11.SEND_RESPONSE:
                     msg = "Server disconnected without sending a response."
                     raise RemoteProtocolError(msg)
 
-                self.h11_state.receive_data(data)
+                self._h11_state.receive_data(data)
             else:
                 assert event is not h11.NEED_DATA
                 break
@@ -186,27 +244,26 @@ class SyncHTTP11Connection(SyncBaseHTTPConnection):
     def _response_closed(self) -> None:
         logger.trace(
             "response_closed our_state=%r their_state=%r",
-            self.h11_state.our_state,
-            self.h11_state.their_state,
+            self._h11_state.our_state,
+            self._h11_state.their_state,
         )
         if (
-            self.h11_state.our_state is h11.DONE
-            and self.h11_state.their_state is h11.DONE
+            self._h11_state.our_state is h11.DONE
+            and self._h11_state.their_state is h11.DONE
         ):
-            self.h11_state.start_next_cycle()
-            self.state = ConnectionState.IDLE
+            self._h11_state.start_next_cycle()
+            self._state = ConnectionState.IDLE
+            if self._keepalive_expiry is not None:
+                self._should_expire_at = self._now() + self._keepalive_expiry
         else:
             self.close()
 
     def close(self) -> None:
-        if self.state != ConnectionState.CLOSED:
-            self.state = ConnectionState.CLOSED
+        if self._state != ConnectionState.CLOSED:
+            self._state = ConnectionState.CLOSED
 
-            if self.h11_state.our_state is h11.MUST_CLOSE:
+            if self._h11_state.our_state is h11.MUST_CLOSE:
                 event = h11.ConnectionClosed()
-                self.h11_state.send(event)
+                self._h11_state.send(event)
 
             self.socket.close()
-
-    def is_socket_readable(self) -> bool:
-        return self.socket.is_readable()
