@@ -1,5 +1,7 @@
+import enum
+import time
 from ssl import SSLContext
-from typing import Iterator, Dict, List, Tuple, cast
+from typing import Iterator, Dict, List, Optional, Tuple, cast
 
 import h2.connection
 import h2.events
@@ -12,12 +14,18 @@ from .._bytestreams import IteratorByteStream
 from .._exceptions import LocalProtocolError, PoolTimeout, RemoteProtocolError
 from .._types import URL, Headers, TimeoutDict
 from .._utils import get_logger
-from .base import SyncByteStream, ConnectionState, NewConnectionRequired
+from .base import SyncByteStream, NewConnectionRequired
 from .http import SyncBaseHTTPConnection
 
 logger = get_logger(__name__)
 
 
+class ConnectionState(enum.IntEnum):
+    IDLE = 0
+    ACTIVE = 1
+    CLOSED = 2
+
+
 class SyncHTTP2Connection(SyncBaseHTTPConnection):
     READ_NUM_BYTES = 64 * 1024
     CONFIG = H2Configuration(validate_inbound_headers=False)
@@ -26,32 +34,77 @@ class SyncHTTP2Connection(SyncBaseHTTPConnection):
         self,
         socket: SyncSocketStream,
         backend: SyncBackend,
-        ssl_context: SSLContext = None,
+        keepalive_expiry: float = None,
     ):
         self.socket = socket
-        self.ssl_context = SSLContext() if ssl_context is None else ssl_context
 
-        self.backend = backend
-        self.h2_state = h2.connection.H2Connection(config=self.CONFIG)
+        self._backend = backend
+        self._h2_state = h2.connection.H2Connection(config=self.CONFIG)
 
-        self.sent_connection_init = False
-        self.streams = {}  # type: Dict[int, SyncHTTP2Stream]
-        self.events = {}  # type: Dict[int, List[h2.events.Event]]
+        self._sent_connection_init = False
+        self._streams: Dict[int, SyncHTTP2Stream] = {}
+        self._events: Dict[int, List[h2.events.Event]] = {}
 
-        self.state = ConnectionState.ACTIVE
+        self._keepalive_expiry: Optional[float] = keepalive_expiry
+        self._should_expire_at: Optional[float] = None
+        self._state = ConnectionState.ACTIVE
+        self._exhausted_available_stream_ids = False
 
     def __repr__(self) -> str:
-        return f"<SyncHTTP2Connection state={self.state}>"
+        return f"<SyncHTTP2Connection [{self._state}]>"
 
     def info(self) -> str:
-        return f"HTTP/2, {self.state.name}, {len(self.streams)} streams"
+        return f"HTTP/2, {self._state.name}, {len(self._streams)} streams"
+
+    def _now(self) -> float:
+        return time.monotonic()
+
+    def should_close(self) -> bool:
+        """
+        Return `True` if the connection is currently idle, and the keepalive
+        timeout has passed.
+        """
+        return (
+            self._state == ConnectionState.IDLE
+            and self._should_expire_at is not None
+            and self._now() >= self._should_expire_at
+        )
+
+    def is_idle(self) -> bool:
+        """
+        Return `True` if the connection is currently idle.
+        """
+        return self._state == ConnectionState.IDLE
+
+    def is_closed(self) -> bool:
+        """
+        Return `True` if the connection has been closed.
+        """
+        return self._state == ConnectionState.CLOSED
+
+    def is_available(self) -> bool:
+        """
+        Return `True` if the connection is currently able to accept an outgoing request.
+        This occurs when any of the following occur:
+
+        * The connection has not yet been opened, and HTTP/2 support is enabled.
+          We don't *know* at this point if we'll end up on an HTTP/2 connection or
+          not, but we *might* do, so we indicate availability.
+        * The connection has been opened, and is currently idle.
+        * The connection is open, and is an HTTP/2 connection. The connection must
+          also not have exhausted the maximum total number of stream IDs.
+        """
+        return (
+            self._state != ConnectionState.CLOSED
+            and not self._exhausted_available_stream_ids
+        )
 
     @property
     def init_lock(self) -> SyncLock:
         # We do this lazily, to make sure backend autodetection always
         # runs within an async context.
         if not hasattr(self, "_initialization_lock"):
-            self._initialization_lock = self.backend.create_lock()
+            self._initialization_lock = self._backend.create_lock()
         return self._initialization_lock
 
     @property
@@ -59,7 +112,7 @@ class SyncHTTP2Connection(SyncBaseHTTPConnection):
         # We do this lazily, to make sure backend autodetection always
         # runs within an async context.
         if not hasattr(self, "_read_lock"):
-            self._read_lock = self.backend.create_lock()
+            self._read_lock = self._backend.create_lock()
         return self._read_lock
 
     @property
@@ -67,24 +120,17 @@ class SyncHTTP2Connection(SyncBaseHTTPConnection):
         # We do this lazily, to make sure backend autodetection always
         # runs within an async context.
         if not hasattr(self, "_max_streams_semaphore"):
-            max_streams = self.h2_state.local_settings.max_concurrent_streams
-            self._max_streams_semaphore = self.backend.create_semaphore(
+            max_streams = self._h2_state.local_settings.max_concurrent_streams
+            self._max_streams_semaphore = self._backend.create_semaphore(
                 max_streams, exc_class=PoolTimeout
             )
         return self._max_streams_semaphore
 
     def start_tls(
-        self, hostname: bytes, timeout: TimeoutDict = None
+        self, hostname: bytes, ssl_context: SSLContext, timeout: TimeoutDict = None
     ) -> SyncSocketStream:
         raise NotImplementedError("TLS upgrade not supported on HTTP/2 connections.")
 
-    def get_state(self) -> ConnectionState:
-        return self.state
-
-    def mark_as_ready(self) -> None:
-        if self.state == ConnectionState.IDLE:
-            self.state = ConnectionState.READY
-
     def handle_request(
         self,
         method: bytes,
@@ -96,25 +142,26 @@ class SyncHTTP2Connection(SyncBaseHTTPConnection):
         timeout = cast(TimeoutDict, extensions.get("timeout", {}))
 
         with self.init_lock:
-            if not self.sent_connection_init:
+            if not self._sent_connection_init:
                 # The very first stream is responsible for initiating the connection.
-                self.state = ConnectionState.ACTIVE
+                self._state = ConnectionState.ACTIVE
                 self.send_connection_init(timeout)
-                self.sent_connection_init = True
+                self._sent_connection_init = True
 
         self.max_streams_semaphore.acquire()
         try:
             try:
-                stream_id = self.h2_state.get_next_available_stream_id()
+                stream_id = self._h2_state.get_next_available_stream_id()
             except NoAvailableStreamIDError:
-                self.state = ConnectionState.FULL
+                self._exhausted_available_stream_ids = True
                 raise NewConnectionRequired()
             else:
-                self.state = ConnectionState.ACTIVE
+                self._state = ConnectionState.ACTIVE
+                self._should_expire_at = None
 
             h2_stream = SyncHTTP2Stream(stream_id=stream_id, connection=self)
-            self.streams[stream_id] = h2_stream
-            self.events[stream_id] = []
+            self._streams[stream_id] = h2_stream
+            self._events[stream_id] = []
             return h2_stream.handle_request(
                 method, url, headers, stream, extensions
             )
@@ -130,7 +177,7 @@ class SyncHTTP2Connection(SyncBaseHTTPConnection):
         # Need to set these manually here instead of manipulating via
         # __setitem__() otherwise the H2Connection will emit SettingsUpdate
         # frames in addition to sending the undesired defaults.
-        self.h2_state.local_settings = Settings(
+        self._h2_state.local_settings = Settings(
             client=True,
             initial_values={
                 # Disable PUSH_PROMISE frames from the server since we don't do anything
@@ -145,14 +192,14 @@ class SyncHTTP2Connection(SyncBaseHTTPConnection):
         # Some websites (*cough* Yahoo *cough*) balk at this setting being
         # present in the initial handshake since it's not defined in the original
         # RFC despite the RFC mandating ignoring settings you don't know about.
-        del self.h2_state.local_settings[
+        del self._h2_state.local_settings[
             h2.settings.SettingCodes.ENABLE_CONNECT_PROTOCOL
         ]
 
         logger.trace("initiate_connection=%r", self)
-        self.h2_state.initiate_connection()
-        self.h2_state.increment_flow_control_window(2 ** 24)
-        data_to_send = self.h2_state.data_to_send()
+        self._h2_state.initiate_connection()
+        self._h2_state.increment_flow_control_window(2 ** 24)
+        data_to_send = self._h2_state.data_to_send()
         self.socket.write(data_to_send, timeout)
 
     def is_socket_readable(self) -> bool:
@@ -160,8 +207,8 @@ class SyncHTTP2Connection(SyncBaseHTTPConnection):
 
     def close(self) -> None:
         logger.trace("close_connection=%r", self)
-        if self.state != ConnectionState.CLOSED:
-            self.state = ConnectionState.CLOSED
+        if self._state != ConnectionState.CLOSED:
+            self._state = ConnectionState.CLOSED
 
             self.socket.close()
 
@@ -172,13 +219,13 @@ class SyncHTTP2Connection(SyncBaseHTTPConnection):
         WindowUpdated frames have increased the flow rate.
         https://tools.ietf.org/html/rfc7540#section-6.9
         """
-        local_flow = self.h2_state.local_flow_control_window(stream_id)
-        connection_flow = self.h2_state.max_outbound_frame_size
+        local_flow = self._h2_state.local_flow_control_window(stream_id)
+        connection_flow = self._h2_state.max_outbound_frame_size
         flow = min(local_flow, connection_flow)
         while flow == 0:
             self.receive_events(timeout)
-            local_flow = self.h2_state.local_flow_control_window(stream_id)
-            connection_flow = self.h2_state.max_outbound_frame_size
+            local_flow = self._h2_state.local_flow_control_window(stream_id)
+            connection_flow = self._h2_state.max_outbound_frame_size
             flow = min(local_flow, connection_flow)
         return flow
 
@@ -191,9 +238,9 @@ class SyncHTTP2Connection(SyncBaseHTTPConnection):
         an event is available.
         """
         with self.read_lock:
-            while not self.events[stream_id]:
+            while not self._events[stream_id]:
                 self.receive_events(timeout)
-        return self.events[stream_id].pop(0)
+        return self._events[stream_id].pop(0)
 
     def receive_events(self, timeout: TimeoutDict) -> None:
         """
@@ -203,7 +250,7 @@ class SyncHTTP2Connection(SyncBaseHTTPConnection):
         if data == b"":
             raise RemoteProtocolError("Server disconnected")
 
-        events = self.h2_state.receive_data(data)
+        events = self._h2_state.receive_data(data)
         for event in events:
             event_stream_id = getattr(event, "stream_id", 0)
             logger.trace("receive_event stream_id=%r event=%s", event_stream_id, event)
@@ -211,53 +258,58 @@ class SyncHTTP2Connection(SyncBaseHTTPConnection):
             if hasattr(event, "error_code"):
                 raise RemoteProtocolError(event)
 
-            if event_stream_id in self.events:
-                self.events[event_stream_id].append(event)
+            if event_stream_id in self._events:
+                self._events[event_stream_id].append(event)
 
-        data_to_send = self.h2_state.data_to_send()
+        data_to_send = self._h2_state.data_to_send()
         self.socket.write(data_to_send, timeout)
 
     def send_headers(
         self, stream_id: int, headers: Headers, end_stream: bool, timeout: TimeoutDict
     ) -> None:
         logger.trace("send_headers stream_id=%r headers=%r", stream_id, headers)
-        self.h2_state.send_headers(stream_id, headers, end_stream=end_stream)
-        self.h2_state.increment_flow_control_window(2 ** 24, stream_id=stream_id)
-        data_to_send = self.h2_state.data_to_send()
+        self._h2_state.send_headers(stream_id, headers, end_stream=end_stream)
+        self._h2_state.increment_flow_control_window(2 ** 24, stream_id=stream_id)
+        data_to_send = self._h2_state.data_to_send()
         self.socket.write(data_to_send, timeout)
 
     def send_data(
         self, stream_id: int, chunk: bytes, timeout: TimeoutDict
     ) -> None:
         logger.trace("send_data stream_id=%r chunk=%r", stream_id, chunk)
-        self.h2_state.send_data(stream_id, chunk)
-        data_to_send = self.h2_state.data_to_send()
+        self._h2_state.send_data(stream_id, chunk)
+        data_to_send = self._h2_state.data_to_send()
         self.socket.write(data_to_send, timeout)
 
     def end_stream(self, stream_id: int, timeout: TimeoutDict) -> None:
         logger.trace("end_stream stream_id=%r", stream_id)
-        self.h2_state.end_stream(stream_id)
-        data_to_send = self.h2_state.data_to_send()
+        self._h2_state.end_stream(stream_id)
+        data_to_send = self._h2_state.data_to_send()
         self.socket.write(data_to_send, timeout)
 
     def acknowledge_received_data(
         self, stream_id: int, amount: int, timeout: TimeoutDict
     ) -> None:
-        self.h2_state.acknowledge_received_data(amount, stream_id)
-        data_to_send = self.h2_state.data_to_send()
+        self._h2_state.acknowledge_received_data(amount, stream_id)
+        data_to_send = self._h2_state.data_to_send()
         self.socket.write(data_to_send, timeout)
 
     def close_stream(self, stream_id: int) -> None:
         try:
             logger.trace("close_stream stream_id=%r", stream_id)
-            del self.streams[stream_id]
-            del self.events[stream_id]
+            del self._streams[stream_id]
+            del self._events[stream_id]
 
-            if not self.streams:
-                if self.state == ConnectionState.ACTIVE:
-                    self.state = ConnectionState.IDLE
-                elif self.state == ConnectionState.FULL:
-                    self.close()
+            if not self._streams:
+                if self._state == ConnectionState.ACTIVE:
+                    if self._exhausted_available_stream_ids:
+                        self.close()
+                    else:
+                        self._state = ConnectionState.IDLE
+                        if self._keepalive_expiry is not None:
+                            self._should_expire_at = (
+                                self._now() + self._keepalive_expiry
+                            )
         finally:
             self.max_streams_semaphore.release()
 
